#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <linux/module.h>

#define LED3 21  // GPIO pin number for LED3 (make sure it's correct)
static dev_t devno;
static struct class *led_class;
static struct cdev led_cdev;
struct file_operations led_fops;

const int first_minor = 0;
const int max_devices = 255;

static int mygpio_init(void) {
    int err = 0;

    // Request GPIO for LED3
    err = gpio_request(LED3, "LED3");
    if (err) {
        pr_err("Failed to request GPIO for LED3\n");
        return err;
    }

    // Set GPIO direction (output)
    err = gpio_direction_output(LED3, 0);  // Initialize to 0 (off)
    if (err) {
        pr_err("Failed to set GPIO direction for LED3\n");
        goto err_free_gpio;
    }

    // Allocate Major/Minor numbers
    err = alloc_chrdev_region(&devno, first_minor, max_devices, "led-driver");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        goto err_free_gpio;
    }

    pr_info("LED driver got Major %i\n", MAJOR(devno));

    // Create class
    led_class = class_create(THIS_MODULE, "led_class");
    if (IS_ERR(led_class)) {
        pr_err("Failed to create class for LED3\n");
        err = PTR_ERR(led_class);
        goto err_unregister_chrdev;
    }

    // Initialize and add cdev
    cdev_init(&led_cdev, &led_fops);
    err = cdev_add(&led_cdev, devno, 1);  // Register one device only
    if (err < 0) {
        pr_err("Failed to add cdev for LED3\n");
        goto err_destroy_class;
    }

    return 0;

err_destroy_class:
    class_destroy(led_class);
err_unregister_chrdev:
    unregister_chrdev_region(devno, max_devices);
err_free_gpio:
    gpio_free(LED3);

    return err;
}

static void mygpio_exit(void) {
    cdev_del(&led_cdev);
    unregister_chrdev_region(devno, max_devices);
    class_destroy(led_class);
    gpio_free(LED3);
}

int mygpio_open(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Opening LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

int mygpio_release(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Closing/Releasing LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

ssize_t mygpio_write(struct file *filep, const char __user *ubuf, size_t count, loff_t *f_pos) {
    int len, value;
    char kbuf[12];

    len = count < sizeof(kbuf) ? count : sizeof(kbuf) - 1;

    if (copy_from_user(kbuf, ubuf, len)) {
        return -EFAULT;
    }

    kbuf[len] = '\0'; // Null-terminate the string

    if (kstrtoint(kbuf, 10, &value)) {
        pr_err("Error converting string to int\n");
        return -EINVAL;
    }

    gpio_set_value(LED3, value);  // Set GPIO value (output)

    pr_info("Wrote %i to LED3\n", value);
    *f_pos += len;
    return len;
}

ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;

    value = gpio_get_value(LED3);  // Read GPIO value (output)
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos > 0) {
        return 0; // EOF
    }

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len;
    return len;
}

struct file_operations led_fops = {
    .open = mygpio_open,
    .release = mygpio_release,
    .read = mygpio_read,
    .write = mygpio_write,
};

module_init(mygpio_init);
module_exit(mygpio_exit);

MODULE_AUTHOR("Khaled <Khaled.uni.au@gmail.com>");
MODULE_LICENSE("GPL");
