#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <linux/module.h>

#define SW2 16  // GPIO pin number for SW2
static dev_t devno;
static struct class *switch_class;
static struct cdev switch_cdev;
struct file_operations switch_fops;

const int first_minor = 0;
const int max_devices = 255;

static int mygpio_init(void) {
    int err = 0;

    // Request GPIO for SW2
    err = gpio_request(SW2, "SW2");
    if (err) {
        pr_err("Failed to request GPIO for SW2\n");
        return err;
    }

    // Set GPIO direction (input)
    err = gpio_direction_input(SW2);
    if (err) {
        pr_err("Failed to set GPIO direction for SW2\n");
        goto err_free_gpio;
    }

    // Allocate Major/Minor numbers
    err = alloc_chrdev_region(&devno, first_minor, max_devices, "switch-driver");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        goto err_free_gpio;
    }

    pr_info("Switch driver got Major %i\n", MAJOR(devno));

    // Create class
    switch_class = class_create(THIS_MODULE, "switch_class");
    if (IS_ERR(switch_class)) {
        pr_err("Failed to create class for SW2\n");
        err = PTR_ERR(switch_class);
        goto err_unregister_chrdev;
    }

    // Initialize and add cdev
    cdev_init(&switch_cdev, &switch_fops);
    err = cdev_add(&switch_cdev, devno, 1);  // Register one device only
    if (err < 0) {
        pr_err("Failed to add cdev for SW2\n");
        goto err_destroy_class;
    }

    return 0;

err_destroy_class:
    class_destroy(switch_class);
err_unregister_chrdev:
    unregister_chrdev_region(devno, max_devices);
err_free_gpio:
    gpio_free(SW2);

    return err;
}

static void mygpio_exit(void) {
    cdev_del(&switch_cdev);
    unregister_chrdev_region(devno, max_devices);
    class_destroy(switch_class);
    gpio_free(SW2);
}

int mygpio_open(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Opening Switch Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

int mygpio_release(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Closing/Releasing Switch Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;

    // Read GPIO value (input)
    value = gpio_get_value(SW2);
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos > 0) {
        return 0; // EOF
    }

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len;
    return len;
}

struct file_operations switch_fops = {
    .open = mygpio_open,
    .release = mygpio_release,
    .read = mygpio_read,
};

module_init(mygpio_init);
module_exit(mygpio_exit);

MODULE_AUTHOR("Khaled <Khaled.uni.au@gmail.com>");
MODULE_LICENSE("GPL");
