#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/of.h>

#define LED3 21  // GPIO pin number for LED3
static dev_t devno;
static struct class *led_class;
static struct cdev led_cdev;
struct file_operations led_fops;

const int first_minor = 0;
const int max_devices = 255;

// Platform driver match table
static const struct of_device_id led_of_match[] = {
    { .compatible = "au-ece, plat_drv" },  // Compatible name used in Device Tree
    {}
};
MODULE_DEVICE_TABLE(of, led_of_match);

// Probe function
static int plat_drv_probe(struct platform_device *pdev) {
    int err;

    // Request GPIO for LED3
    err = gpio_request(LED3,"LED3");
    if (err) {
        pr_err("Failed to request GPIO for LED3\n");
        return err;
    }

    // Set GPIO direction (output)
    err = gpio_direction_output(LED3, 0);  // Initialize to 0 (off)
    if (err) {
        pr_err("Failed to set GPIO direction for LED3\n");
        gpio_free(LED3);
        return err;
    }

    // Allocate Major/Minor numbers
    err = alloc_chrdev_region(&devno, first_minor, max_devices, "led-driver");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        gpio_free(LED3);
        return err;
    }

    pr_info("LED driver got Major %i\n", MAJOR(devno));

    // Create class
    led_class = class_create(THIS_MODULE,"led_class");
    if (IS_ERR(led_class)) {
        pr_err("Failed to create class for LED3\n");
        err = PTR_ERR(led_class);
        unregister_chrdev_region(devno, max_devices);
        gpio_free(LED3);
        return err;
    }

    // Initialize and add cdev
    cdev_init(&led_cdev, &led_fops);
    err = cdev_add(&led_cdev, devno, 1);  // Register one device only
    if (err < 0) {
        pr_err("Failed to add cdev for LED3\n");
        class_destroy(led_class);
        unregister_chrdev_region(devno, max_devices);
        gpio_free(LED3);
        return err;
    }

    // Create device node in /dev/
    device_create(led_class, NULL, devno, NULL, "plat_drv");

    pr_info("hello from probe\n");

    return 0;
}

// Remove function
static int plat_drv_remove(struct platform_device *pdev) {
    pr_info("goodbye from remove\n");

    device_destroy(led_class, devno);
    cdev_del(&led_cdev);
    class_destroy(led_class);
    unregister_chrdev_region(devno, max_devices);
    gpio_free(LED3);

    return 0;
}

// Platform driver structure
static struct platform_driver led_platform_driver = {
    .probe = plat_drv_probe,
    .remove = plat_drv_remove,
    .driver = {
        .name = "plat_drv",
        .of_match_table = led_of_match,
        .owner = THIS_MODULE,
    },
};

// Init function (register platform driver)
static int __init mygpio_init(void) {
    pr_info("Initializing LED Platform Driver\n");
    return platform_driver_register(&led_platform_driver);
}

// Exit function (unregister platform driver)
static void __exit mygpio_exit(void) {
    platform_driver_unregister(&led_platform_driver);
}

// File operations functions
int mygpio_open(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Opening LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

int mygpio_release(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Closing/Releasing LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

ssize_t mygpio_write(struct file *filep, const char __user *ubuf, size_t count, loff_t *f_pos) {
    int len, value;
    char kbuf[12];

    len = count < sizeof(kbuf) ? count : sizeof(kbuf) - 1;

    if (copy_from_user(kbuf, ubuf, len)) {
        return -EFAULT;
    }

    kbuf[len] = '\0'; // Null-terminate the string

    if (kstrtoint(kbuf, 10, &value)) {
        pr_err("Error converting string to int\n");
        return -EINVAL;
    }

    gpio_set_value(LED3, value);  // Set GPIO value (output)

    pr_info("Wrote %i to LED3\n", value);
    *f_pos += len;
    return len;
}

ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;

    value = gpio_get_value(LED3);  // Read GPIO value (output)
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos > 0) {
        return 0; // EOF
    }

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len;
    return len;
}

struct file_operations led_fops = {
    .open = mygpio_open,
    .release = mygpio_release,
    .read = mygpio_read,
    .write = mygpio_write,
};

module_init(mygpio_init);
module_exit(mygpio_exit);

MODULE_AUTHOR("Khaled <Khaled.uni.au@gmail.com>");
MODULE_LICENSE("GPL");
