#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/of.h>

#define NUM_GPIOS 2  // Antal GPIO'er

// Struktur til at gemme GPIO-oplysninger
struct gpio_dev {
    int no;   // GPIO nummer
    int dir;  // 0: input, 1: output
};

// Array der definerer GPIO numre og retninger
static struct gpio_dev gpio_devs[NUM_GPIOS] = {
    {12, 0},  // GPIO12, input
    {21, 1}   // GPIO21, output
};
static int gpios_len = 2;

static dev_t devno;
static struct class *led_class;
static struct cdev led_cdev;
struct file_operations led_fops;

const int first_minor = 0;
const int max_devices = 255;

// Platform driver match table
static const struct of_device_id led_of_match[] = {
    { .compatible = "au-ece, plat_drv" },
    {}
};
MODULE_DEVICE_TABLE(of, led_of_match);

// Probe function
static int plat_drv_probe(struct platform_device *pdev) {
    int err, i;
    dev_t curr_devno;

    // Allocate Major/Minor numbers
    err = alloc_chrdev_region(&devno, first_minor, max_devices, "led-driver");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        return err;
    }
    pr_info("LED driver got Major %i\n", MAJOR(devno));

    // Create class
    led_class = class_create(THIS_MODULE, "led_class");
    if (IS_ERR(led_class)) {
        pr_err("Failed to create class\n");
        err = PTR_ERR(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
    }

    // Initialize and add cdev
    cdev_init(&led_cdev, &led_fops);
    err = cdev_add(&led_cdev, devno, NUM_GPIOS);  // Registrer to devices
    if (err < 0) {
        pr_err("Failed to add cdev\n");
        class_destroy(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
    }

    // Loop over gpio_devs array og anmod om GPIO'er samt sæt deres retning
    for (i = 0; i < NUM_GPIOS; i++) {
        struct gpio_dev *dev = &gpio_devs[i];
        
        // Anmod om GPIO
        err = gpio_request(dev->no, "gpio_dev");
        if (err) {
            pr_err("Failed to request GPIO %d\n", dev->no);
            goto cleanup_gpio;
        }

        // Sæt GPIO retning
        if (dev->dir == 1) {  // Hvis det er output
            err = gpio_direction_output(dev->no, 0);  // Initialiser til 0
        } else {  // Hvis det er input
            err = gpio_direction_input(dev->no);
        }
        if (err) {
            pr_err("Failed to set GPIO direction for GPIO %d\n", dev->no);
            gpio_free(dev->no);
            goto cleanup_gpio;
        }

        // Opret device node for hver GPIO
        curr_devno = MKDEV(MAJOR(devno), MINOR(devno) + i);
        device_create(led_class, NULL, curr_devno, NULL, "plat_drv%d", i);
        pr_info("Created device for GPIO %d with minor %d\n", dev->no, MINOR(curr_devno));
    }

    pr_info("hello from probe\n");
    return 0;

    cleanup_gpio:
        while (i--) {
            gpio_free(gpio_devs[i].no);
            device_destroy(led_class, MKDEV(MAJOR(devno), MINOR(devno) + i));
        }
        cdev_del(&led_cdev);
        class_destroy(led_class);
        unregister_chrdev_region(devno, max_devices);
        return err;
}

// Remove function
static int plat_drv_remove(struct platform_device *pdev) {
    int i;
    pr_info("goodbye from remove\n");

    // Fjern devices, cdev og class
    for (i = 0; i < NUM_GPIOS; i++) {
        gpio_free(gpio_devs[i].no);
        device_destroy(led_class, MKDEV(MAJOR(devno), MINOR(devno) + i));
    }
    cdev_del(&led_cdev);
    class_destroy(led_class);
    unregister_chrdev_region(devno, max_devices);

    return 0;
}

// Platform driver structure
static struct platform_driver led_platform_driver = {
    .probe = plat_drv_probe,
    .remove = plat_drv_remove,
    .driver = {
        .name = "plat_drv",
        .of_match_table = led_of_match,
        .owner = THIS_MODULE,
    },
};

// Init function (register platform driver)
static int __init mygpio_init(void) {
    pr_info("Initializing LED Platform Driver\n");
    return platform_driver_register(&led_platform_driver);
}

// Exit function (unregister platform driver)
static void __exit mygpio_exit(void) {
    platform_driver_unregister(&led_platform_driver);
}

// File operations functions
int mygpio_open(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Opening LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

int mygpio_release(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Closing/Releasing LED Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

ssize_t mygpio_write(struct file *filep, const char __user *ubuf, size_t count, loff_t *f_pos) {
    int len, value;
    char kbuf[12];
    int minor = MINOR(filep->f_inode->i_rdev);  // Få minor nummer
    struct gpio_dev *dev;

    // Tjek om minor nummeret er gyldigt
    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];  // Få den korrekte gpio_dev struct baseret på minor nummeret

    // Kontroller, om GPIO'en er sat til output
    if (dev->dir != 1) {
        pr_err("GPIO %d is not configured as output\n", dev->no);
        return -EINVAL;
    }

    len = count < sizeof(kbuf) ? count : sizeof(kbuf) - 1;

    if (copy_from_user(kbuf, ubuf, len)) {
        return -EFAULT;
    }

    kbuf[len] = '\0';  // Null-terminér strengen

    if (kstrtoint(kbuf, 10, &value)) {
        pr_err("Error converting string to int\n");
        return -EINVAL;
    }

    gpio_set_value(dev->no, value);  // Sæt GPIO-værdien (output)

    pr_info("Wrote %i to GPIO %d\n", value, dev->no);
    *f_pos += len;
    return len;
}

ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;
    int minor = MINOR(filep->f_inode->i_rdev);  // Få minor nummer
    struct gpio_dev *dev;

    // Tjek om minor nummeret er gyldigt
    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];  // Få den korrekte gpio_dev struct baseret på minor nummeret

    // Kontroller, om GPIO'en er sat til input
    if (dev->dir != 0) {
        pr_err("GPIO %d is not configured as input\n", dev->no);
        return -EINVAL;
    }

    value = gpio_get_value(dev->no);  // Læs GPIO-værdien (input)
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos > 0) {
        return 0;  // EOF
    }

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len;
    return len;
}

struct file_operations led_fops = {
    .open = mygpio_open,
    .release = mygpio_release,
    .read = mygpio_read,
    .write = mygpio_write,
};

module_init(mygpio_init);
module_exit(mygpio_exit);

MODULE_AUTHOR("Khaled <Khaled.uni.au@gmail.com>");
MODULE_LICENSE("GPL");
