#include <linux/gpio.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/uaccess.h>
#include <linux/module.h>
#include <linux/of.h>
#include <linux/platform_device.h>
#include <linux/of_gpio.h>
#include <linux/err.h>

#define NUM_GPIOS 21  // Antal GPIO'er (tilpasses til dynamisk senere)

// Struktur til at gemme GPIO-oplysninger
struct gpio_dev {
    int no;   // GPIO nummer
    int dir;  // Retning
    int toggle_state;  // 1=toggling, 0=ikke toggling
    int toggle_delay;  // Delay i ms
    struct timer_list timer;  // Kernel-timer
};

static struct gpio_dev gpio_devs[NUM_GPIOS];
static int gpios_len;
static dev_t devno;
static struct class *gpio_class; // Omdøbt fra 'led_class' til 'gpio_class'
static struct cdev gpio_cdev; // Omdøbt fra 'led_cdev' til 'gpio_cdev'
struct file_operations gpio_fops; // Omdøbt fra 'led_fops' til 'gpio_fops'

// Match-table for Device Tree
static const struct of_device_id gpio_of_match[] = {
    { .compatible = "mygpio, plat_drv" },
    {},
};
MODULE_DEVICE_TABLE(of, gpio_of_match);

// Timer callback
static void gpio_timer_callback(struct timer_list *t) {
    struct gpio_dev *dev = from_timer(dev, t, timer);

    // Hvis toggling er aktiv for denne GPIO
    if (dev->toggle_state) {
        // Skift GPIO-tilstand
        gpio_set_value(dev->no, !gpio_get_value(dev->no));

        // Planlæg næste toggling baseret på delay
        mod_timer(&dev->timer, jiffies + msecs_to_jiffies(dev->toggle_delay));
    }
}


// Probe function for the platform driver
static int plat_drv_probe(struct platform_device *pdev) {
    int err, i;
    dev_t curr_devno;
    struct device *dev = &pdev->dev;
    struct device_node *np = dev->of_node;
    enum of_gpio_flags flag;

    pr_info("Probing platform driver\n");

    // Retrieve the number of GPIOs from the Device Tree
    gpios_len = of_gpio_count(np);
    if (gpios_len <= 0) {
        pr_err("No GPIOs defined in Device Tree\n");
        return -EINVAL;
    }

    pr_info("Found %d GPIOs in Device Tree\n", gpios_len);

    // Loop through GPIOs in the Device Tree and retrieve configurations
    for (i = 0; i < gpios_len; i++) {
        gpio_devs[i].no = of_get_gpio(np, i);
        if (gpio_devs[i].no < 0) {
            pr_err("Failed to get GPIO %d\n", i);
            return gpio_devs[i].no;
        }

        if (of_get_gpio_flags(np, i, &flag) < 0) {
            pr_err("Failed to get GPIO flags for GPIO %d\n", gpio_devs[i].no);
            return -EINVAL;
        }

        // Set GPIO direction: 0 = input, 1 = output
        gpio_devs[i].dir = (flag == OF_GPIO_ACTIVE_LOW) ? 0 : 1;

        pr_info("Configured GPIO %d with direction %d\n", gpio_devs[i].no, gpio_devs[i].dir);
    }

    // Create character devices and set up GPIO configurations
    for (i = 0; i < gpios_len; i++) {
        struct gpio_dev *dev = &gpio_devs[i];

        // Initialize toggling parameters
        dev->toggle_state = 0;         // Default: not toggling
        dev->toggle_delay = 1000;      // Default delay in ms
        timer_setup(&dev->timer, gpio_timer_callback, 0); // Initialize kernel timer

        // Request the GPIO
        err = gpio_request(dev->no, "gpio_dev");
        if (err) {
            pr_err("Failed to request GPIO %d\n", dev->no);
            goto cleanup_gpio;
        }

        // Set GPIO direction
        if (dev->dir == 1) { // output
            err = gpio_direction_output(dev->no, 0);
        } else { // input
            err = gpio_direction_input(dev->no);
        }
        if (err) {
            pr_err("Failed to set direction for GPIO %d\n", dev->no);
            gpio_free(dev->no);
            goto cleanup_gpio;
        }

        // Create a device node for each GPIO
        curr_devno = MKDEV(MAJOR(devno), MINOR(devno) + i);
        device_create(gpio_class, NULL, curr_devno, &gpio_devs[i], "plat_drv%d", i);
        pr_info("Created device for GPIO %d with minor %d\n", dev->no, MINOR(curr_devno));
    }

    pr_info("Driver successfully probed with GPIO configuration from Device Tree\n");
    return 0;

cleanup_gpio:
    // Cleanup in case of failure
    while (i--) {
        del_timer_sync(&gpio_devs[i].timer); // Remove the timer
        gpio_free(gpio_devs[i].no);          // Free the GPIO
        device_destroy(gpio_class, MKDEV(MAJOR(devno), MINOR(devno) + i)); // Destroy the device node
    }
    return err;
}



// Remove function
static int plat_drv_remove(struct platform_device *pdev) {
    int i;

    for (i = 0; i < gpios_len; i++) {
        struct gpio_dev *dev = &gpio_devs[i];

        del_timer_sync(&dev->timer);  // Stop timer
        gpio_free(dev->no);           // Frigiv GPIO
        device_destroy(gpio_class, MKDEV(MAJOR(devno), MINOR(devno) + i));
    }

    cdev_del(&gpio_cdev);
    class_destroy(gpio_class);
    unregister_chrdev_region(devno, gpios_len);

    return 0;
}


// Platform driver structure
static struct platform_driver gpio_platform_driver = {
    .probe = plat_drv_probe,
    .remove = plat_drv_remove,
    .driver = {
        .name = "plat_drv",
        .of_match_table = gpio_of_match, // Matcher med Device Tree
        .owner = THIS_MODULE,
    },
};

// gpio_toggle_state show:
static ssize_t gpio_toggle_state_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    if (!gpio_dev) return -EINVAL;

    return sprintf(buf, "%d\n", gpio_dev->toggle_state);
}

// gpio_toggle_state store:
static ssize_t gpio_toggle_state_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    int state;

    if (!gpio_dev) return -EINVAL;

    if (kstrtoint(buf, 10, &state)) return -EINVAL;

    gpio_dev->toggle_state = state;

    if (state) {
        // Start timer with the current delay
        mod_timer(&gpio_dev->timer, jiffies + msecs_to_jiffies(gpio_dev->toggle_delay));
    } else {
        // Stop timer
        del_timer_sync(&gpio_dev->timer);
    }

    return size;
}



// gpio_toggle_delay show 
static ssize_t gpio_toggle_delay_show(struct device *dev, struct device_attribute *attr, char *buf) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    if (!gpio_dev) return -EINVAL;

    return sprintf(buf, "%d\n", gpio_dev->toggle_delay);
}

// gpio_toggle_delay store
static ssize_t gpio_toggle_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t size) {
    struct gpio_dev *gpio_dev = dev_get_drvdata(dev);
    int delay;

    if (!gpio_dev) return -EINVAL;

    if (kstrtoint(buf, 10, &delay)) return -EINVAL;

    gpio_dev->toggle_delay = delay;
    return size;
}


DEVICE_ATTR_RW( gpio_toggle_delay );
DEVICE_ATTR_RW( gpio_toggle_state );

// Attributter til SysFS:
static struct attribute *gpio_attrs[] = {
    &dev_attr_gpio_toggle_state.attr,
    &dev_attr_gpio_toggle_delay.attr,
    NULL,
};
ATTRIBUTE_GROUPS( gpio );


// Init function (register platform driver)
static int __init mygpio_init(void) {
    int err;

    pr_info("Initializing GPIO Platform Driver\n");

    // Alloker chrdev region
    err = alloc_chrdev_region(&devno, 0, NUM_GPIOS, "plat_drv");
    if (err < 0) {
        pr_err("Failed to allocate char dev region\n");
        return err;
    }

    // Opret class og cdev
    gpio_class = class_create(THIS_MODULE, "plat_drv_class");
    if (IS_ERR(gpio_class)) {
        unregister_chrdev_region(devno, NUM_GPIOS);
        return PTR_ERR(gpio_class);
    }
    gpio_class->dev_groups = gpio_groups;

    cdev_init(&gpio_cdev, &gpio_fops);
    gpio_cdev.owner = THIS_MODULE;
    err = cdev_add(&gpio_cdev, devno, NUM_GPIOS);
    if (err) {
        class_destroy(gpio_class);
        unregister_chrdev_region(devno, NUM_GPIOS);
        return err;
    }

    return platform_driver_register(&gpio_platform_driver);
}

// Exit function (unregister platform driver)
static void __exit mygpio_exit(void) {
    platform_driver_unregister(&gpio_platform_driver);
    cdev_del(&gpio_cdev);
    class_destroy(gpio_class);
    unregister_chrdev_region(devno, NUM_GPIOS);
}

// File operations functions
int mygpio_open(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Opening GPIO Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

int mygpio_release(struct inode *inode, struct file *filep) {
    int major, minor;
    major = MAJOR(inode->i_rdev);
    minor = MINOR(inode->i_rdev);
    pr_info("Closing/Releasing GPIO Device [major], [minor]: %i, %i\n", major, minor);
    return 0;
}

ssize_t mygpio_write(struct file *filep, const char __user *ubuf, size_t count, loff_t *f_pos) {
    int len, value;
    char kbuf[12];
    int minor = MINOR(filep->f_inode->i_rdev);
    struct gpio_dev *dev;

    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];
    if (dev->dir != 1) {
        pr_err("GPIO %d is not configured as output\n", dev->no);
        return -EINVAL;
    }

    len = count < sizeof(kbuf) ? count : sizeof(kbuf) - 1;
    if (copy_from_user(kbuf, ubuf, len)) {
        return -EFAULT;
    }

    kbuf[len] = '\0';
    if (kstrtoint(kbuf, 10, &value)) {
        pr_err("Error converting string to int\n");
        return -EINVAL;
    }

    gpio_set_value(dev->no, value);
    pr_info("Wrote %i to GPIO %d\n", value, dev->no);
    *f_pos += len;
    return len;
}
ssize_t mygpio_read(struct file *filep, char __user *buf, size_t count, loff_t *f_pos) {
    int value;
    char kbuf[12];
    int len;
    int minor = MINOR(filep->f_inode->i_rdev);
    struct gpio_dev *dev;

    if (minor >= gpios_len) {
        pr_err("Invalid minor number\n");
        return -ENODEV;
    }

    dev = &gpio_devs[minor];
    value = gpio_get_value(dev->no);
    len = snprintf(kbuf, sizeof(kbuf), "%d\n", value);

    if (*f_pos >= len) // Handle EOF
        return 0;

    if (copy_to_user(buf, kbuf, len)) {
        return -EFAULT;
    }

    *f_pos += len; // Advance position
    return len;
}

// File operations
struct file_operations gpio_fops = {
    .owner = THIS_MODULE,
    .open = mygpio_open,
    .release = mygpio_release,
    .write = mygpio_write,
    .read = mygpio_read,
};

module_init(mygpio_init);
module_exit(mygpio_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Dit Navn");
MODULE_DESCRIPTION("GPIO Platform Driver");
